from django.db.models import Exists, OuterRef

from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo

from corpstats.models import CorpStat

field_label = 'Corporation Stats'

scopes = [
    'esi-corporations.track_members.v1',
    'esi-universe.read_structures.v1'
]

permissions = ['corpstats.add_corpstat']


def add_character(request, token):
    if EveCharacter.objects.filter(character_id=token.character_id).exists():
        corp_id = EveCharacter.objects.get(character_id=token.character_id).corporation_id
    else:
        corp_id = \
            token.get_esi_client().Character.get_characters_character_id(
                character_id=token.character_id).result()['corporation_id']
    try:
        corp = EveCorporationInfo.objects.get(corporation_id=corp_id)
    except EveCorporationInfo.DoesNotExist:
        corp = EveCorporationInfo.objects.create_corporation(corp_id)
    cs = CorpStat.objects.create(token=token, corp=corp)
    cs.update()
    assert cs.pk  # ensure update was successful


def is_character_added(character: EveCharacter):
    return (
        CorpStat.objects
        .filter(token__character_id=character.character_id)
        .exists()
    )


is_character_added_annotation = Exists(
    CorpStat.objects
    .filter(token__character_id=OuterRef('character_id'))
)



from django import forms

from .app_imports import import_apps
from .app_settings import CHARLINK_IGNORE_APPS


class LinkForm(forms.Form):
    add_character = forms.BooleanField(
        required=False,
        initial=True,
        disabled=True,
        label='Add Character (default)'
    )

    def __init__(self, user, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for app, data in import_apps().items():
            if app != 'add_character' and app not in CHARLINK_IGNORE_APPS and user.has_perms(data['permissions']):
                self.fields[app] = forms.BooleanField(
                    required=False,
                    initial=True,
                    label=data.get('field_label', app)
                )



from django import template

from allianceauth.eveonline.models import EveCorporationInfo, EveCharacter

register = template.Library()


@register.filter
def get_corp_members(corp: EveCorporationInfo):
    return EveCharacter.objects.filter(corporation_id=corp.corporation_id).select_related("character_ownership__user__profile__main_character")


@register.filter
def get_char_attr(character_obj, attr: str):
    if isinstance(character_obj, str):
        try:
            character_obj = int(character_obj)
        except:
            return ''

    if isinstance(character_obj, EveCharacter):
        return getattr(character_obj, attr)
    elif isinstance(character_obj, int):
        try:
            return getattr(EveCharacter.objects.get(pk=character_obj), attr)
        except:
            return ''
    else:
        return ''



from django.db import models


class General(models.Model):
    class Meta:
        managed = False
        default_permissions = ()
        permissions = (
            ('view_corp', 'Can view linked character of members of their corporation.'),
            ('view_alliance', 'Can view linked character of members of their alliance.'),
            ('view_state', 'Can view linked character of members of their auth state.'),
        )



from django.db import transaction
from django.db.models import Exists, OuterRef

from miningtaxes.models import Character
from miningtaxes import tasks

from allianceauth.eveonline.models import EveCharacter

field_label = "Mining Taxes"

scopes = Character.get_esi_scopes()

permissions = ["miningtaxes.basic_access"]


def add_character(request, token):
    eve_character = EveCharacter.objects.get(character_id=token.character_id)
    with transaction.atomic():
        character, _ = Character.objects.update_or_create(eve_character=eve_character)
    tasks.update_character.delay(character_pk=character.pk)


def is_character_added(character: EveCharacter):
    return Character.objects.filter(eve_character=character).exists()


is_character_added_annotation = Exists(
    Character.objects
    .filter(eve_character_id=OuterRef('pk'))
)



from django.db.models import Exists, OuterRef

from corptools.models import CharacterAudit
from corptools.tasks import update_character
from corptools.app_settings import get_character_scopes, CORPTOOLS_APP_NAME

from allianceauth.eveonline.models import EveCharacter

field_label = CORPTOOLS_APP_NAME

scopes = get_character_scopes()

permissions = []


def add_character(request, token):
    CharacterAudit.objects.update_or_create(
        character=EveCharacter.objects.get_character_by_id(token.character_id))
    update_character.apply_async(args=[token.character_id], priority=6)


def is_character_added(character: EveCharacter):
    return CharacterAudit.objects.filter(character=character).exists()


is_character_added_annotation = Exists(
    CharacterAudit.objects
    .filter(character_id=OuterRef('pk'))
)



from django.db.models import Exists, OuterRef

from moonmining.models import Owner
from moonmining import __title__, tasks
from moonmining.app_settings import MOONMINING_ADMIN_NOTIFICATIONS_ENABLED

from app_utils.messages import messages_plus
from app_utils.allianceauth import notify_admins

from allianceauth.eveonline.models import EveCorporationInfo, EveCharacter

field_label = __title__

scopes = Owner.esi_scopes()

permissions = ["moonmining.add_refinery_owner", "moonmining.basic_access"]


def add_character(request, token):
    character_ownership = request.user.character_ownerships.select_related(
        "character"
    ).get(character__character_id=token.character_id)

    try:
        corporation = EveCorporationInfo.objects.get(
            corporation_id=character_ownership.character.corporation_id
        )
    except EveCorporationInfo.DoesNotExist:
        corporation = EveCorporationInfo.objects.create_corporation(
            corp_id=character_ownership.character.corporation_id
        )
        corporation.save()

    owner, _ = Owner.objects.update_or_create(
        corporation=corporation,
        defaults={"character_ownership": character_ownership},
    )
    tasks.update_owner.delay(owner.pk)
    messages_plus.success(request, f"Update of refineres started for {owner}.")
    if MOONMINING_ADMIN_NOTIFICATIONS_ENABLED:
        notify_admins(
            message=("%(corporation)s was added as new owner by %(user)s.")
            % {"corporation": owner, "user": request.user},
            title=f"{__title__}: Owner added: {owner}",
        )


def is_character_added(character: EveCharacter):
    return Owner.objects.filter(
        character_ownership__character=character
    ).exists()


is_character_added_annotation = Exists(
    Owner.objects
    .filter(character_ownership__character_id=OuterRef('pk'))
)



from django.shortcuts import render, redirect, get_object_or_404
from django.contrib.auth.decorators import login_required
from django.contrib.auth.models import User, Permission
from django.contrib import messages
from django.db.models import Exists, OuterRef, Q
from django.core.exceptions import PermissionDenied
from django.http import Http404

from allianceauth.services.hooks import get_extension_logger
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from allianceauth.authentication.decorators import permissions_required
from allianceauth.authentication.models import CharacterOwnership

from app_utils.allianceauth import users_with_permission

from .forms import LinkForm
from .app_imports import import_apps
from .decorators import charlink
from .app_settings import CHARLINK_IGNORE_APPS

logger = get_extension_logger(__name__)


def get_visible_corps(user: User):
    char = user.profile.main_character

    corps = EveCorporationInfo.objects.filter(
        Exists(
            CharacterOwnership.objects
            .filter(character__corporation_id=OuterRef('corporation_id'))
        )
    )

    if user.is_superuser:
        corps = corps.all()
    else:
        queries = []
        has_access = False

        if user.has_perm('charlink.view_alliance'):
            queries.append(Q(alliance__alliance_id=char.alliance_id))
            has_access = True

        if user.has_perm('charlink.view_corp') and not user.has_perm('charlink.view_alliance'):
            queries.append(Q(corporation_id=char.corporation_id))
            has_access = True

        if user.has_perm('charlink.view_state'):
            alliances = user.profile.state.member_alliances.all()
            corporations = user.profile.state.member_corporations.all()

            queries.append(
                Q(alliance__alliance_id__in=alliances.values('alliance_id')) |
                Q(id__in=corporations)
            )
            has_access = True

        if not has_access:
            query = queries.pop()
            for q in queries:
                query |= q

            corps = corps.filter(query)
        else:
            corps = corps.none()

    return corps


def chars_annotate_linked_apps(characters, apps: dict):
    for app, data in apps.items():
        characters = characters.annotate(
            **{app: data['is_character_added_annotation']}
        )

    return characters


def get_user_available_apps(user: User):
    imported_apps = import_apps()

    return {
        app: data
        for app, data in imported_apps.items()
        if app not in CHARLINK_IGNORE_APPS and user.has_perms(data['permissions'])
    }


def get_user_linked_chars(user: User):
    available_apps = get_user_available_apps(user)

    return {
        'apps': available_apps,
        'characters': chars_annotate_linked_apps(
            EveCharacter.objects.filter(character_ownership__user=user),
            available_apps
        )
    }


@login_required
def index(request):
    imported_apps = import_apps()

    if request.method == 'POST':
        form = LinkForm(request.user, request.POST)
        if form.is_valid():

            scopes = set()
            selected_apps = []

            for app, to_import in form.cleaned_data.items():
                if to_import:
                    scopes.update(imported_apps[app].get('scopes', []))
                    selected_apps.append(app)

            logger.debug(f"Scopes: {scopes}")

            request.session['charlink'] = {
                'scopes': list(scopes),
                'apps': selected_apps,
            }

            return redirect('charlink:login')

    else:
        form = LinkForm(request.user)

    context = {
        'form': form,
        'characters_added': get_user_linked_chars(request.user),
        'is_auditor': request.user.has_perm('charlink.view_state') or request.user.has_perm('charlink.view_corp') or request.user.has_perm('charlink.view_alliance'),
    }

    return render(request, 'charlink/charlink.html', context=context)


@login_required
@charlink
def login_view(request, token):
    imported_apps = import_apps()

    charlink_data = request.session.pop('charlink')

    for app in charlink_data['apps']:
        if app != 'add_character' and app not in CHARLINK_IGNORE_APPS and request.user.has_perms(imported_apps[app]['permissions']):
            try:
                imported_apps[app]['add_character'](request, token)
            except Exception as e:
                logger.exception(e)
                messages.error(request, f"Failed to add character to {imported_apps[app]['field_label']}")
            else:
                messages.success(request, f"Character successfully added to {imported_apps[app]['field_label']}")

    return redirect('charlink:index')


@login_required
@permissions_required([
    'charlink.view_corp',
    'charlink.view_alliance',
    'charlink.view_state',
])
def audit(request, corp_id=None):
    char = request.user.profile.main_character
    if corp_id:
        corp = get_object_or_404(EveCorporationInfo, corporation_id=corp_id)
    else:
        corp = None

    corps = get_visible_corps(request.user)

    if corp_id and corp not in corps:
        raise PermissionDenied('You do not have permission to view the selected corporation statistics.')

    if not corp_id and corps.count() == 1:
        corp = corps.first()
    elif not corp_id and char:
        try:
            corp = corps.get(corporation_id=char.corporation_id)
        except EveCorporationInfo.DoesNotExist:
            pass

    context = {
        'available': corps,
        'selected': corp,
        'available_apps': get_user_available_apps(request.user),
    }

    return render(request, 'charlink/audit.html', context=context)


@login_required
@permissions_required([
    'charlink.view_corp',
    'charlink.view_alliance',
    'charlink.view_state',
])
def search(request):
    search_string = request.GET.get('search_string', None)
    if not search_string:
        return redirect('charlink:index')

    corps = get_visible_corps(request.user)

    characters = (
        EveCharacter.objects
        .filter(
            character_name__icontains=search_string,
            corporation_id__in=corps.values('corporation_id'),
        )
        .order_by('character_name')
        .select_related('character_ownership__user__profile__main_character')
    )

    context = {
        'search_string': search_string,
        'characters': characters,
        'available': corps,
        'available_apps': get_user_available_apps(request.user),
    }

    return render(request, 'charlink/search.html', context=context)


@login_required
@permissions_required([
    'charlink.view_corp',
    'charlink.view_alliance',
    'charlink.view_state',
])
def audit_user(request, user_id):
    user = get_object_or_404(User, pk=user_id)

    corps = get_visible_corps(request.user)

    if (
        not request.user.is_superuser
        and
        user != request.user
        and
        not corps
        .filter(
            corporation_id=user.profile.main_character.corporation_id
        )
        .exists()
    ):
        raise PermissionDenied('You do not have permission to view the selected user statistics.')

    context = {
        'characters_added': get_user_linked_chars(user),
        'available': corps,
        'available_apps': get_user_available_apps(user),
    }

    return render(request, 'charlink/user_audit.html', context=context)


@login_required
@permissions_required([
    'charlink.view_corp',
    'charlink.view_alliance',
    'charlink.view_state',
])
def audit_app(request, app):
    imported_apps = import_apps()

    if app not in imported_apps:
        raise Http404()

    app_data = imported_apps[app]

    if not request.user.has_perms(app_data['permissions']):
        raise PermissionDenied('You do not have permission to view the selected application statistics.')

    corps = get_visible_corps(request.user)

    queries = []
    for perm in app_data['permissions']:
        app_label, codename = perm.split('.')
        perm_obj = Permission.objects.get(content_type__app_label=app_label, codename=codename)

        queries.append(Q(character_ownership__user__in=users_with_permission(perm_obj)))

    if len(queries) == 0:
        perm_query = Q(character_ownership__isnull=False)
    else:
        perm_query = queries.pop()
        for query in queries:
            perm_query |= query

    visible_characters = EveCharacter.objects.filter(
        (
            Q(corporation_id__in=corps.values('corporation_id')) |
            Q(character_ownership__user__profile__main_character__corporation_id__in=corps.values('corporation_id'))
        ) &
        perm_query,
    ).select_related('character_ownership__user__profile__main_character')

    visible_characters = chars_annotate_linked_apps(
        visible_characters,
        {app: app_data}
    ).order_by(app, 'character_name')

    context = {
        'characters': visible_characters,
        'available': corps,
        'app': app,
        'app_data': app_data,
        'available_apps': get_user_available_apps(request.user),
    }

    return render(request, 'charlink/app_audit.html', context=context)



from django.db import transaction
from django.db.models import Exists, OuterRef

from memberaudit.models import Character, ComplianceGroupDesignation
from memberaudit.app_settings import MEMBERAUDIT_APP_NAME, MEMBERAUDIT_TASKS_NORMAL_PRIORITY
from memberaudit import tasks

from allianceauth.eveonline.models import EveCharacter

field_label = MEMBERAUDIT_APP_NAME

scopes = Character.get_esi_scopes()

permissions = ["memberaudit.basic_access"]


def add_character(request, token):
    eve_character = EveCharacter.objects.get(character_id=token.character_id)
    with transaction.atomic():
        character, created = Character.objects.update_or_create(
            eve_character=eve_character, defaults={"is_disabled": False}
        )
    tasks.update_character.apply_async(
        kwargs={"character_pk": character.pk},
        priority=MEMBERAUDIT_TASKS_NORMAL_PRIORITY,
    )
    if ComplianceGroupDesignation.objects.exists():
        tasks.update_compliance_groups_for_user.apply_async(
            args=[request.user.pk], priority=MEMBERAUDIT_TASKS_NORMAL_PRIORITY
        )


def is_character_added(character: EveCharacter):
    return Character.objects.filter(eve_character=character).exists()


is_character_added_annotation = Exists(
    Character.objects
    .filter(eve_character_id=OuterRef('pk'))
)



from django.db.models import Exists, OuterRef

from moonstuff.providers import ESI_CHARACTER_SCOPES
from moonstuff.models import TrackingCharacter
from moonstuff.tasks import import_extraction_data

from allianceauth.eveonline.models import EveCharacter

field_label = 'Moon Tools'

scopes = ESI_CHARACTER_SCOPES

permissions = ['moonstuff.add_trackingcharacter']


def add_character(request, token):
    eve_char = EveCharacter.objects.get(character_id=token.character_id)

    if not TrackingCharacter.objects.filter(character=eve_char).exists():
        char = TrackingCharacter(character=eve_char)
        char.save()

        # Schedule an import task to pull data from the new Tracking Character.
        import_extraction_data.delay()


def is_character_added(character: EveCharacter):
    return TrackingCharacter.objects.filter(character=character).exists()


is_character_added_annotation = Exists(
    TrackingCharacter.objects
    .filter(character_id=OuterRef('pk'))
)



from importlib import import_module

from django.conf import settings
from django.db.models import Exists, OuterRef

from allianceauth.services.hooks import get_extension_logger
from allianceauth.authentication.models import CharacterOwnership

logger = get_extension_logger(__name__)

_supported_apps = {
    'add_character': {
        'field_label': 'Add Character (default)',
        'add_character': lambda request, token: None,
        'scopes': ['publicData'],
        'permissions': [],
        'is_character_added': lambda character: CharacterOwnership.objects.filter(character=character).exists(),
        'is_character_added_annotation': Exists(CharacterOwnership.objects.filter(character_id=OuterRef('pk')))
    }
}

_imported = False


def import_apps():
    global _imported
    if not _imported:
        for app in settings.INSTALLED_APPS:
            if app != 'allianceauth':
                try:
                    module = import_module(f'charlink.imports.{app}')
                except ModuleNotFoundError:
                    logger.debug(f"Loading of {app} link: failed")
                else:
                    _supported_apps[app] = {
                        'field_label': module.field_label,
                        'add_character': module.add_character,
                        'scopes': module.scopes,
                        'permissions': module.permissions,
                        'is_character_added': module.is_character_added,
                        'is_character_added_annotation': module.is_character_added_annotation
                    }

                    logger.debug(f"Loading of {app} link: success")

        _imported = True

    return _supported_apps
